"""System Bridge GUI: Main"""
from __future__ import annotations

import asyncio
import json
import os
import sys
from typing import Optional

from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QApplication, QMessageBox
import async_timeout
from systembridgeshared.base import Base
from systembridgeshared.const import SETTING_LOG_LEVEL
from systembridgeshared.database import Database
from systembridgeshared.exceptions import (
    AuthenticationException,
    ConnectionErrorException,
)
from systembridgeshared.logger import setup_logger
from systembridgeshared.models.media_play import MediaPlay
from systembridgeshared.models.notification import Notification
from systembridgeshared.settings import Settings
from systembridgeshared.websocket_client import WebSocketClient
from typer import Typer

from ._version import __version__
from .system_tray import SystemTray
from .widgets.timed_message_box import TimedMessageBox
from .window.main import MainWindow
from .window.notification import NotificationWindow
from .window.player import PlayerWindow


class Main(Base):
    """Main"""

    def __init__(
        self,
        command: str = "main",
        gui_only: bool = False,
        data: Optional[dict] = None,
    ) -> None:
        """Initialize"""
        super().__init__()
        self._logger.info("System Bridge GUI %s: Startup", __version__.public())

        self._database = database
        self._settings = settings

        self._application = QApplication([])
        self._icon = QIcon(os.path.join(os.path.dirname(__file__), "icon.png"))
        self._application.setStyleSheet(
            """
            QWidget {
                color: #FFFFFF;
                background-color: #212121;
            }

            QMenu {
                background-color: #292929;
            }

            QMenu::item {
                background-color: transparent;
            }

            QMenu::item:selected {
                background-color: #757575;
            }
            """
        )

        if command == "main":
            self._logger.info("Main: Setup")

            self._gui_only = gui_only
            self._websocket_client = WebSocketClient(self._settings)

            asyncio.run(self._setup_websocket())
            asyncio.run(self._websocket_client.close())

            self._main_window = MainWindow(
                self._settings,
                self._icon,
            )

            self._system_tray_icon = SystemTray(
                self._database,
                self._settings,
                self._icon,
                self._application,
                self._callback_exit_application,
                self._callback_show_window,
            )
            self._system_tray_icon.show()
        elif command == "media-player-audio":
            self._logger.info("Media Player: Audio")
            if data is None:
                self._logger.error("No data provided!")
                self._startup_error("No data provided!")
                sys.exit(1)
            media_play = MediaPlay(**data)
            self._player_window = PlayerWindow(
                self._settings,
                self._icon,
                self._application,
                "audio",
                media_play,
            )
        elif command == "media-player-video":
            self._logger.info("Media Player: Video")
            if data is None:
                self._logger.error("No data provided!")
                self._startup_error("No data provided!")
                sys.exit(1)
            media_play = MediaPlay(**data)
            self._player_window = PlayerWindow(
                self._settings,
                self._icon,
                self._application,
                "video",
                media_play,
            )
        elif command == "notification":
            self._logger.info("Notification")
            if data is None:
                self._logger.error("No data provided!")
                self._startup_error("No data provided!")
                sys.exit(1)
            self._player_window = NotificationWindow(
                self._settings,
                self._icon,
                self._application,
                Notification(**data),
            )

        sys.exit(self._application.exec())

    def _callback_exit_application(self) -> None:
        """Exit the application"""
        asyncio.run(self._exit_application(self._gui_only))

    def _callback_show_window(
        self,
        path: str,
        maximized: bool,
        width: Optional[int] = 1280,
        height: Optional[int] = 720,
    ) -> None:
        """Show the main window"""
        self._logger.info("Showing window: %s", path)

        if width is None:
            width = 1280
        if height is None:
            height = 720

        self._main_window.hide()
        self._main_window.setup(path)
        self._main_window.resize(width, height)
        screen_geometry = self._application.primaryScreen().availableSize()
        self._main_window.move(
            int((screen_geometry.width() - self._main_window.width()) / 2),
            int((screen_geometry.height() - self._main_window.height()) / 2),
        )
        if maximized:
            self._main_window.showMaximized()
        else:
            self._main_window.showNormal()

    def _startup_error(
        self,
        message: str,
    ) -> None:
        """Handle a startup error"""
        error_message = TimedMessageBox(
            10,
            f"{message} Exiting in",
        )
        error_message.setIcon(QMessageBox.Critical)
        error_message.setWindowTitle("Error")
        error_message.exec()
        # Exit cleanly
        self._logger.info("Exit GUI..")
        self._application.quit()
        sys.exit(1)

    async def _exit_application(
        self,
        gui_only: bool,
    ) -> None:
        """Exit the backend"""
        if not gui_only:
            self._logger.info("Exit Backend..")
            await self._setup_websocket()
            await self._websocket_client.exit_backend()
            await self._websocket_client.close()
        self._logger.info("Exit GUI..")
        self._application.quit()

    async def _setup_websocket(self) -> None:
        """Setup the WebSocket client"""
        try:
            async with async_timeout.timeout(20):
                await self._websocket_client.connect()
        except AuthenticationException as exception:
            self._logger.error("Authentication failed: %s", exception)
            self._startup_error("Authentication failed!")
        except ConnectionErrorException as exception:
            self._logger.error("Could not connect to WebSocket: %s", exception)
            self._startup_error("Could not connect to WebSocket!")
        except asyncio.TimeoutError as exception:
            self._logger.error("Connection timeout to WebSocket: %s", exception)
            self._startup_error("Connection timeout to WebSocket!")


app = Typer()


@app.command(name="main", help="Run the main application")
def main(
    gui_only: bool = False,
) -> None:
    """Run the main application"""
    Main(command="main", gui_only=gui_only)


@app.command(name="media-player", help="Run the media player")
def media_player(
    media_type: str,
    data: str,
) -> None:
    """Run the media player"""
    Main(command=f"media-player-{media_type}", data=json.loads(data))


@app.command(name="notification", help="Show a notification")
def notification(
    data: str,
) -> None:
    """Show a notification"""
    Main(command="notification", data=json.loads(data))


if __name__ == "__main__":
    asyncio.set_event_loop(asyncio.new_event_loop())
    database = Database()
    settings = Settings(database)

    log_level: str = str(settings.get(SETTING_LOG_LEVEL))

    setup_logger(log_level, "system-bridge-gui")

    app()



"""System Bridge GUI: Player Window"""
from json import dumps
import sys
from urllib.parse import urlencode

from PySide6 import QtCore
from PySide6.QtCore import QUrl
from PySide6.QtGui import QIcon, Qt
from PySide6.QtWebEngineWidgets import QWebEngineView
from PySide6.QtWidgets import QApplication, QFrame, QVBoxLayout
from systembridgeshared.base import Base
from systembridgeshared.const import (
    QUERY_API_KEY,
    QUERY_API_PORT,
    SECRET_API_KEY,
    SETTING_PORT_API,
)
from systembridgeshared.models.notification import Notification
from systembridgeshared.settings import Settings


class NotificationWindow(Base, QFrame):
    """Notification Window"""

    def __init__(
        self,
        settings: Settings,
        icon: QIcon,
        application: QApplication,
        notification: Notification,
    ) -> None:
        """Initialize the window"""
        Base.__init__(self)
        QFrame.__init__(
            self,
            WindowFlags=Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint,  # type: ignore
        )

        self._settings = settings

        self.layout = QVBoxLayout(self)  # type: ignore
        self.layout.setContentsMargins(0, 0, 0, 0)  # type: ignore

        self.browser = QWebEngineView()

        self.layout.addWidget(self.browser)  # type: ignore

        self.setWindowIcon(icon)

        height: int = 48
        title_lines: int = 1 + int(round(len(notification.title) / 52, 0))
        self._logger.info("Title Lines: %s", title_lines)
        if title_lines > 1:
            height += 64 * title_lines
        if notification.message is not None:
            height += 24
            message_lines: int = 1 + int(round(len(notification.message) / 62, 0))
            self._logger.info("Message Lines: %s", message_lines)
            if message_lines > 1:
                height += 20 * message_lines
        if notification.image is not None:
            height += 280
        if notification.actions is not None and len(notification.actions) > 0:
            height += 72

        self._logger.info("Height: %s", height)

        self.resize(420, height)

        screen_geometry = application.primaryScreen().availableSize()

        self.move(
            screen_geometry.width() - self.width() - 8,
            screen_geometry.height() - self.height() - 8,
        )

        notification_dict = notification.dict(exclude_none=True)
        if "actions" in notification_dict:
            # Fix encoding issue by converting array to json
            notification_dict["actions"] = dumps(notification_dict["actions"])

        api_port = self._settings.get(SETTING_PORT_API)
        api_key = self._settings.get_secret(SECRET_API_KEY)
        url = QUrl(
            f"""http://localhost:{api_port}/app/notification.html?{urlencode({
                    QUERY_API_KEY: api_key,
                    QUERY_API_PORT: api_port,
                    **notification_dict,
                })}"""
        )
        self._logger.info("Open URL: %s", url)
        self.browser.load(url)

        self.browser.urlChanged.connect(self._url_changed)  # type: ignore

        if notification.timeout is None or notification.timeout < 1:
            notification.timeout = 5
        self.time_to_wait = int(notification.timeout)
        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(1000)
        self.timer.timeout.connect(self._timer_changed)  # type: ignore

        self.showNormal()

        self.timer.start()

    def _timer_changed(self):
        """Change the content of the message box"""
        self.time_to_wait -= 1
        if self.time_to_wait < 0:
            self.close()
            sys.exit(0)

    def _url_changed(self, url: QUrl):
        """Handle URL changes"""
        self._logger.info("URL Changed: %s", url)
        if url.host() == "close.window":
            self._logger.info("Close Window Requested. Closing Window.")
            self.close()
            sys.exit(0)



"""System Bridge GUI: Player Window"""
import sys
from urllib.parse import urlencode

from PySide6.QtCore import QUrl
from PySide6.QtGui import QIcon, Qt
from PySide6.QtWebEngineWidgets import QWebEngineView
from PySide6.QtWidgets import QApplication, QFrame, QVBoxLayout
from systembridgeshared.base import Base
from systembridgeshared.const import (
    QUERY_API_KEY,
    QUERY_API_PORT,
    SECRET_API_KEY,
    SETTING_PORT_API,
)
from systembridgeshared.models.media_play import MediaPlay
from systembridgeshared.settings import Settings


class PlayerWindow(Base, QFrame):
    """Player Window"""

    def __init__(
        self,
        settings: Settings,
        icon: QIcon,
        application: QApplication,
        media_type: str,
        media_play: MediaPlay,
    ) -> None:
        """Initialize the window"""
        Base.__init__(self)
        QFrame.__init__(
            self,
            WindowFlags=Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint,  # type: ignore
        )

        self._settings = settings

        self.layout = QVBoxLayout(self)  # type: ignore
        self.layout.setContentsMargins(0, 0, 0, 0)  # type: ignore

        self.browser = QWebEngineView()

        self.layout.addWidget(self.browser)  # type: ignore

        self.setWindowTitle("System Bridge - Player")
        self.setWindowIcon(icon)

        if media_type == "audio":
            self.resize(540, 140)
        elif media_type == "video":
            self.resize(480, 270)

        screen_geometry = application.primaryScreen().availableSize()

        self.move(
            screen_geometry.width() - self.width() - 8,
            screen_geometry.height() - self.height() - 8,
        )

        api_port = self._settings.get(SETTING_PORT_API)
        api_key = self._settings.get_secret(SECRET_API_KEY)
        url = QUrl(
            f"""http://localhost:{api_port}/app/player/{media_type}.html?{urlencode({
                    QUERY_API_KEY: api_key,
                    QUERY_API_PORT: api_port,
                    **media_play.dict(exclude_none=True),
                })}"""
        )
        self._logger.info("Open URL: %s", url)
        self.browser.load(url)

        self.browser.urlChanged.connect(self._url_changed)  # type: ignore

        self.showNormal()

    def _url_changed(self, url: QUrl):
        """Handle URL changes"""
        self._logger.info("URL Changed: %s", url)
        if url.host() == "close.window":
            self._logger.info("Close Window Requested. Closing Window.")
            self.close()
            sys.exit(0)



"""System Bridge GUI: System Tray"""
from __future__ import annotations

from collections.abc import Callable
import os
from typing import Optional
from webbrowser import open_new_tab

from PySide6.QtGui import QAction, QCursor, QIcon
from PySide6.QtWidgets import QApplication, QMenu, QSystemTrayIcon
from pyperclip import copy
from systembridgeshared.base import Base
from systembridgeshared.common import get_user_data_directory
from systembridgeshared.database import Database
from systembridgeshared.models.database_data import System as DatabaseSystem
from systembridgeshared.settings import Settings

PATH_BRIDGES_OPEN_ON = "/app/bridges/openon.html"
PATH_BRIDGES_SETUP = "/app/bridges/setup.html"
PATH_DATA = "/app/data.html"
PATH_SETTINGS = "/app/settings.html"

URL_DISCUSSIONS = "https://github.com/timmo001/system-bridge/discussions"
URL_DOCS = "https://system-bridge.timmo.dev"
URL_ISSUES = "https://github.com/timmo001/system-bridge/issues/new/choose"
URL_LATEST_RELEASE = "https://github.com/timmo001/system-bridge/releases/latest"


class SystemTray(Base, QSystemTrayIcon):
    """System Tray"""

    # pylint: disable=unsubscriptable-object
    def __init__(
        self,
        database: Database,
        settings: Settings,
        icon: QIcon,
        parent: QApplication,
        callback_exit_application: Callable,
        callback_show_window: Callable[[str, bool, Optional[int], Optional[int]], None],
    ) -> None:
        """Initialize the system tray"""
        Base.__init__(self)
        QSystemTrayIcon.__init__(self, icon, parent)

        self._database = database
        self._settings = settings

        self._logger.info("Setup system tray")

        self.callback_show_window = callback_show_window

        self.activated.connect(self._on_activated)  # type: ignore

        menu = QMenu()

        action_settings: QAction = menu.addAction("Open Settings")
        action_settings.triggered.connect(self._show_settings)  # type: ignore

        action_data: QAction = menu.addAction("View Data")
        action_data.triggered.connect(self._show_data)  # type: ignore

        # menu.addSeparator()

        # action_bridges_setup = menu.addAction("Setup Bridges")
        # action_bridges_setup.triggered.connect(self._show_bridges_setup)

        # action_bridges_sendto = menu.addAction("Open URL On..")
        # action_bridges_sendto.triggered.connect(self._show_bridges_send_to)

        menu.addSeparator()

        latest_version_text = "Latest Version"
        result_version_current = self._database.get_data_item_by_key(
            DatabaseSystem, "version"
        )
        version_current = (
            result_version_current.value if result_version_current is not None else None
        )
        result_version_latest = self._database.get_data_item_by_key(
            DatabaseSystem, "version_latest"
        )
        version_latest = (
            result_version_latest.value if result_version_latest is not None else None
        )
        result_version_newer_available = self._database.get_data_item_by_key(
            DatabaseSystem, "version_newer_available"
        )
        version_newer_available: str = (
            result_version_newer_available.value
            if result_version_newer_available is not None
            and result_version_newer_available.value is not None
            else ""
        )

        if version_newer_available.lower() == "true":
            latest_version_text = f"{version_latest} (New)"
        else:
            latest_version_text += f" ({version_current})"

        action_latest_release: QAction = menu.addAction(latest_version_text)
        action_latest_release.triggered.connect(self._open_latest_releases)  # type: ignore

        menu_help = menu.addMenu("Help")

        action_docs: QAction = menu_help.addAction("Documentation / Website")
        action_docs.triggered.connect(self._open_docs)  # type: ignore

        action_feature: QAction = menu_help.addAction("Suggest a Feature")
        action_feature.triggered.connect(self._open_feature_request)  # type: ignore

        action_issue: QAction = menu_help.addAction("Report an issue")
        action_issue.triggered.connect(self._open_issues)  # type: ignore

        action_discussions: QAction = menu_help.addAction("Discussions")
        action_discussions.triggered.connect(self._open_discussions)  # type: ignore

        menu_help.addSeparator()

        action_api_key: QAction = menu_help.addAction("Copy API key to clipboard")
        action_api_key.triggered.connect(self._copy_api_key)  # type: ignore

        menu_help.addSeparator()

        action_log: QAction = menu_help.addAction("Open Log File")
        action_log.triggered.connect(self._open_log)  # type: ignore

        action_log_gui: QAction = menu_help.addAction("Open GUI Log File")
        action_log_gui.triggered.connect(self._open_gui_log)  # type: ignore

        menu.addSeparator()

        action_exit: QAction = menu.addAction("Exit")
        action_exit.triggered.connect(callback_exit_application)  # type: ignore

        self.setContextMenu(menu)

    def _on_activated(
        self,
        reason: int,
    ) -> None:
        """Handle the activated signal"""
        if reason == QSystemTrayIcon.Trigger:
            self.contextMenu().popup(QCursor.pos())

    def _copy_api_key(self) -> None:
        """Copy API key to clipboard"""
        self._logger.info("Copy API key to clipboard")
        key = self._settings.get_secret("api_key")
        copy(key)

    def _open_latest_releases(self) -> None:
        """Open latest release"""
        self._logger.info("Open: %s", URL_LATEST_RELEASE)
        open_new_tab(URL_LATEST_RELEASE)

    def _open_docs(self) -> None:
        """Open documentation"""
        self._logger.info("Open: %s", URL_DOCS)
        open_new_tab(URL_DOCS)

    def _open_feature_request(self) -> None:
        """Open feature request"""
        self._logger.info("Open: %s", URL_ISSUES)
        open_new_tab(URL_ISSUES)

    def _open_issues(self) -> None:
        """Open issues"""
        self._logger.info("Open: %s", URL_ISSUES)
        open_new_tab(URL_ISSUES)

    def _open_discussions(self) -> None:
        """Open discussions"""
        self._logger.info("Open: %s", URL_DISCUSSIONS)
        open_new_tab(URL_DISCUSSIONS)

    def _open_log(self) -> None:
        """Open log"""
        log_path = os.path.join(get_user_data_directory(), "system-bridge.log")
        self._logger.info("Open: %s", log_path)
        open_new_tab(log_path)

    def _open_gui_log(self) -> None:
        """Open GUI log"""
        log_path = os.path.join(get_user_data_directory(), "system-bridge-gui.log")
        self._logger.info("Open: %s", log_path)
        open_new_tab(log_path)

    # def _show_bridges_send_to(self) -> None:
    #     """Show bridges open url on window"""
    #     self.callback_show_window(PATH_BRIDGES_OPEN_ON, False, 620, 420)

    # def _show_bridges_setup(self) -> None:
    #     """Show bridges setup window"""
    #     self.callback_show_window(PATH_BRIDGES_SETUP, False)

    def _show_data(self) -> None:
        """Show api data"""
        self.callback_show_window(PATH_DATA, False)  # type: ignore

    def _show_settings(self) -> None:
        """Show settings"""
        self.callback_show_window(PATH_SETTINGS, False)  # type: ignore
